from abc import abstractmethod, ABC
from typing import Union, List

from .. import OpenGLGUIFacade
from gui.layer import  LayerGroup, Layer, GridLayer, CharactersLayer, TextLayer, UILayer


class OpenGLLayerGroup(LayerGroup):

    def __init__(self, gui: OpenGLGUIFacade):
        self.__gui = gui
        self.__layers = []  # type: List[Union[None,Layer]]
        self.__translationX = 0.0  # type: float
        self.__translationY = 0.0  # type: float

    def __len__(self) -> int:
        return len(self.__layers)

    def __getitem__(self, item) -> Union[None, Layer]:
        return self.__layers[item]

    def __iter__(self):
        return iter(self.__layers)

    # Create layers

    def __addLayer(self, layer: Layer, level):
        if level < 0:
            self.__layers.append(layer)
        elif level < len(self.__layers):
            self.__layers[level] = layer
        else:
            self.__layers.extend([None for l in range(len(self.__layers), level + 1)])
            self.__layers[level] = layer

    def createGridLayer(self, level: int = -1) -> GridLayer:
        from gui.impl.opengl.layer.OpenGLGridLayer import OpenGLGridLayer
        layer = OpenGLGridLayer(self.__gui)
        self.__addLayer(layer, level)
        return layer

    def createCharactersLayer(self, level: int = -1) -> CharactersLayer:
        from gui.impl.opengl.layer import OpenGLCharactersLayer
        layer = OpenGLCharactersLayer(self.__gui)
        self.__addLayer(layer, level)
        return layer

    def createTextLayer(self, level: int = -1) -> TextLayer:
        from gui.impl.opengl.layer.OpenGLTextLayer import OpenGLTextLayer
        layer = OpenGLTextLayer(self.__gui)
        self.__addLayer(layer, level)
        return layer

    def createUILayer(self, level: int = -1) -> UILayer:
        from gui.impl.opengl.layer.OpenGLUILayer import OpenGLUILayer
        layer = OpenGLUILayer(self.__gui)
        self.__addLayer(layer, level)
        return layer

    # Get & set layers

    def getLevelCount(self) -> int:
        return len(self.__layers)

    def getLayer(self, level: int) -> Layer:
        if level < 0 or level >= len(self.__layers):
            raise ValueError("Invalid level {}".format(level))
        return self.__layers[level]

    def moveLayer(self, layer: Union[int, Layer], level: int):
        if type(layer) == Layer:
            layer = self.__layers.index(layer)
        if layer < 0 or layer >= len(self.__layers):
            raise ValueError("Invalid level {}".format(layer))
        if level < 0 or level >= len(self.__layers):
            raise ValueError("Invalid level {}".format(level))
        if level != layer:
            self.__layers[level] = self.__layers[layer]
            self.__layers[layer] = None

    def removeLayer(self, layer: Union[int, Layer]):
        if type(layer) == Layer:
            layer = self.__layers.index(layer)
        if layer < 0 or layer >= len(self.__layers):
            raise ValueError("Invalid level {}".format(layer))
        self.__layers[layer] = None

    # Properties

    @property
    def translationX(self) -> float:
        return self.__translationX

    @property
    def translationY(self) -> float:
        return self.__translationY

    def getTranslation(self) -> (float, float):
        return -self.__translationX / self.__gui.screenPixelWidth, \
               self.__translationY / self.__gui.screenPixelHeight

    def setTranslation(self, x: float, y: float):
        self.__translationX = -x * self.__gui.screenPixelWidth
        self.__translationY = y * self.__gui.screenPixelHeight

