from typing import Union, List

import numpy as np
import typing

from gui.layer import TextLayer
from gui.text import TextStyle, TextToken, ContentTextToken, StyleTextToken, TextTokenizer, TextTokenType
from gui.impl.opengl import OpenGLGUIFacade
from gui.impl.pygame.text import TextTileset
from .OpenGLLayer import OpenGLLayer


class OpenGLTextLayer(OpenGLLayer, TextLayer):

    def __init__(self, gui: OpenGLGUIFacade):
        super(OpenGLTextLayer, self).__init__(gui)
        self.__tileset = TextTileset()
        self.__baseStyle = None  # type: Union[None, TextStyle]

    def updateDepths(self, depth: int, index: int, maxIndex: int) -> int:
        self.setDepths(depth - index)
        return depth - index - 1

    @property
    def maxCharacterCount(self) -> int:
        return self.faceCount

    def setStyle(self, style: TextStyle):
        self.__baseStyle = style

    def setMaxCharacterCount(self, maxCount: int):
        self.createFaces(maxCount)
        self.sendData(dynamic=True)

    def setDisplayedCharacterCount(self, count: int):
        self.setRenderedFaceCount(min(count, self.maxCharacterCount))

    def __updateTileset(self, characters: str):
        # Add characters to the tileset
        if self.__tileset.addCharacters(characters):
            # Send the tileset to the GPU
            self.setTileset(self.__tileset.toNumpyArray())
            if False:  # Save generated tileset (to debug/illustrate)
                from PIL import Image
                if not hasattr(self, "tilesetUpdates"):
                    self.tilesetUpdates = 0
                self.tilesetUpdates += 1
                Image.fromarray(self.__tileset.toNumpyArray()).save("charactersTileset{}.png".format(self.tilesetUpdates))

    def setText(self, x: float, y: float, html: str):
        tokenizer = TextTokenizer(self.__baseStyle)
        tokens = tokenizer.parse(html)
        self.setTextTokens(x, y, tokens)

    def setTextTokens(self, x: float, y: float, tokens: List[TextToken]):
        # First make sure that all character are in the tileset
        for token in tokens:
            if token.type == TextTokenType.CONTENT:
                contentTextToken = typing.cast(ContentTextToken, token)
                content = contentTextToken.content
                self.__updateTileset(content)
            elif token.type == TextTokenType.STYLE:
                styleTextToken = typing.cast(StyleTextToken, token)
                self.__tileset.setStyle(styleTextToken.style)
            else:
                raise RuntimeError("Unsupported token type")

        # Set the faces
        x0 = x
        charIndex = 0
        for token in tokens:
            if token.type == TextTokenType.CONTENT:
                contentTextToken = typing.cast(ContentTextToken, token)
                for character in contentTextToken.content:
                    if character == "\n":
                        x = x0
                        style = self.__tileset.getStyle()
                        y += style.tileHeight
                        continue
                    tile = self.__tileset.getCharacterTile(character)
                    self.setFacePixelTexture(charIndex, tile.x1, tile.y1, tile.x2, tile.y2)
                    self.setFacePixelLocation(charIndex, x, y, x + tile.width, y + tile.height)
                    x += tile.advance
                    charIndex += 1
                    if charIndex >= self.maxCharacterCount:
                        break
            elif token.type == TextTokenType.STYLE:
                styleTextToken = typing.cast(StyleTextToken, token)
                self.__tileset.setStyle(styleTextToken.style)
            else:
                raise RuntimeError("Unsupported token type")
            if charIndex >= self.maxCharacterCount:
                break

        self.setRenderedFaceCount(charIndex)
