from typing import List, Union, Tuple

import numpy as np
from OpenGL.GL import GL_QUADS, glGenVertexArrays, glBindVertexArray, \
    glGenBuffers, glBindBuffer, GL_ARRAY_BUFFER, glBufferData, \
    GL_STATIC_DRAW, glVertexAttribPointer, GL_FLOAT, GL_FALSE, GL_ELEMENT_ARRAY_BUFFER, \
    glEnableVertexAttribArray, glDrawElements, GL_UNSIGNED_INT, glDisableVertexAttribArray, glGenTextures, \
    GL_UNPACK_ALIGNMENT, glPixelStorei, \
    GL_TEXTURE_2D, \
    glBindTexture, glTexParameteri, GL_TEXTURE_BASE_LEVEL, GL_TEXTURE_MAX_LEVEL, glTexImage2D, GL_UNSIGNED_BYTE, \
    GL_RGBA, GL_TEXTURE_MIN_FILTER, GL_TEXTURE_MAG_FILTER, GL_NEAREST, glDeleteBuffers, \
    glDeleteVertexArrays, glDeleteTextures, glBufferSubData, GL_DYNAMIC_DRAW
from PIL import Image

from gui.impl.opengl import OpenGLGUIFacade


class Mesh:

    def __init__(self, gui: OpenGLGUIFacade):
        self.__gui = gui

        # Texture properties
        self.__textureId = -1  # type: int
        self.__textureWidth = 1  # type: int
        self.__textureHeight = 1  # type: int
        self.__texturePixelWidth = 0  # type: float
        self.__texturePixelHeight = 0  # type: float

        # Mesh properties
        self.__vaoId = -1  # type: int
        self.__vboIds = []  # type: List[int]
        self.__faceCount = 0  # type: int
        self.__renderedFaceCount = 0  # type: int

        # Tile properties
        self.__tileWidth = 0  # type: int
        self.__tileHeight = 0  # type: int
        self.__textureTileWidth = 0  # type: float
        self.__textureTileHeight = 0  # type: float
        self.__screenTileWidth = 0  # type: float
        self.__screenTileHeight = 0  # type: float

    @property
    def gui(self) -> OpenGLGUIFacade:
        return self.__gui

    @property
    def screenPixelWidth(self) -> float:
        return self.__gui.screenPixelWidth

    @property
    def screenPixelHeight(self) -> float:
        return self.__gui.screenPixelHeight

    @property
    def textureWidth(self) -> int:
        return self.__textureWidth

    @property
    def textureHeight(self) -> int:
        return self.__textureHeight

    @property
    def texturePixelWidth(self) -> float:
        return self.__texturePixelWidth

    @property
    def texturePixelHeight(self) -> float:
        return self.__texturePixelHeight

    @property
    def tileWidth(self) -> int:
        return self.__tileWidth

    @property
    def tileHeight(self) -> int:
        return self.__tileHeight

    @property
    def screenTileWidth(self) -> float:
        return self.__screenTileWidth

    @property
    def screenTileHeight(self) -> float:
        return self.__screenTileHeight

    @property
    def textureTileWidth(self) -> float:
        return self.__textureTileWidth

    @property
    def textureTileHeight(self) -> float:
        return self.__textureTileHeight

    def setTileSize(self, tileWidth: int, tileHeight: int):
        self.__tileWidth = tileWidth
        self.__tileHeight = tileHeight
        self.__textureTileWidth = float(tileWidth) / float(self.textureWidth)
        self.__textureTileHeight = float(tileHeight) / float(self.textureHeight)
        self.__texturePixelWidth = 1.0 / float(self.textureWidth)
        self.__texturePixelHeight = 1.0 / float(self.textureHeight)
        self.__screenTileWidth = 2 * float(tileWidth) / float(self.gui.screenWidth)
        self.__screenTileHeight = 2 * float(tileHeight) / float(self.gui.screenHeight)

    def setRenderedFaceCount(self, count: int):
        self.__renderedFaceCount = count

    def setTileset(self, image: Union[str, np.array]):

        if type(image) == str:
            # Load texture image
            image = Image.open(image)
            assert image.mode == "RGBA"
            imageArray = np.array(image)
        else:
            imageArray = image
        assert imageArray.ndim == 3
        assert imageArray.shape[2] == 4

        # Create texture from image
        if self.__textureId < 0:
            self.__textureId = glGenTextures(1)
        glPixelStorei(GL_UNPACK_ALIGNMENT, 4)
        glBindTexture(GL_TEXTURE_2D, self.__textureId)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_BASE_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAX_LEVEL, 0)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST)
        glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, imageArray.shape[0], imageArray.shape[1],
                     0, GL_RGBA, GL_UNSIGNED_BYTE, imageArray)

        self.__textureWidth = imageArray.shape[0]
        self.__textureHeight = imageArray.shape[1]

    def setData(self, vertices: np.ndarray, faces: np.ndarray, uvMap: np.ndarray, dynamic: bool = False):
        assert vertices.dtype == np.float32
        assert faces.dtype == np.uint
        assert uvMap.dtype == np.float32

        if vertices.ndim == 4:
            gridWidth = vertices.shape[0]
            gridHeight = vertices.shape[1]
            assert vertices.shape[2] == 4
            assert vertices.shape[3] == 3

            assert faces.ndim == 3
            assert faces.shape[0] == gridWidth
            assert faces.shape[1] == gridHeight
            assert faces.shape[2] == 4

            assert uvMap.ndim == 4
            assert uvMap.shape[0] == gridWidth
            assert uvMap.shape[1] == gridHeight
            assert uvMap.shape[2] == 4
            assert uvMap.shape[3] == 2

            self.__faceCount = gridWidth * gridHeight
        elif vertices.ndim == 3:
            count = vertices.shape[0]
            assert vertices.shape[1] == 4
            assert vertices.shape[2] == 3

            assert faces.ndim == 2
            assert faces.shape[0] == count
            assert faces.shape[1] == 4

            assert uvMap.ndim == 3
            assert uvMap.shape[0] == count
            assert uvMap.shape[1] == 4
            assert uvMap.shape[2] == 2

            self.__faceCount = count
        else:
            raise ValueError("Unsupported input shape")

        self.__renderedFaceCount = self.__faceCount

        # Create one Vertex Array Object (VAO)
        if self.__vaoId == -1:
            self.__vaoId = glGenVertexArrays(1)
            # We will be working on this VAO
            glBindVertexArray(self.__vaoId)

            # Create three Vertex Buffer Objects (VBO)
            self.__vboIds = glGenBuffers(3)

        # Commodity variables to memorize the id of each VBO
        vertexVboId = self.__vboIds[0]
        uvVboId = self.__vboIds[1]
        indexVboId = self.__vboIds[2]

        # Static/Dynamic flag
        if dynamic:
            flags = GL_DYNAMIC_DRAW
        else:
            flags = GL_STATIC_DRAW

        # Copy vertices data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, vertexVboId)
        vertices = np.ascontiguousarray(vertices.flatten())
        glBufferData(GL_ARRAY_BUFFER, 4 * len(vertices), vertices, flags)
        glVertexAttribPointer(0, 3, GL_FLOAT, GL_FALSE, 0, None)

        # Copy UV coords data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, uvVboId)
        uvMap = np.ascontiguousarray(uvMap.flatten())
        glBufferData(GL_ARRAY_BUFFER, 4 * len(uvMap), uvMap, flags)
        glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 0, None)

        # Copy index data to GPU
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVboId)
        faces = np.ascontiguousarray(faces.flatten())
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 4 * len(faces), faces, flags)

    def updateData(self, vertices: np.ndarray, faces: np.ndarray, uvMap: np.ndarray):
        if self.__vaoId == -1 or vertices.shape[0] != self.__faceCount:
            return self.setData(vertices, faces, uvMap, dynamic=True)

        assert vertices.dtype == np.float32
        assert faces.dtype == np.uint
        assert uvMap.dtype == np.float32

        if vertices.ndim == 4:
            gridWidth = vertices.shape[0]
            gridHeight = vertices.shape[1]
            assert (gridWidth * gridHeight) == self.__faceCount
            assert vertices.shape[2] == 4
            assert vertices.shape[3] == 3

            assert faces.ndim == 3
            assert faces.shape[0] == gridWidth
            assert faces.shape[1] == gridHeight
            assert faces.shape[2] == 4

            assert uvMap.ndim == 4
            assert uvMap.shape[0] == gridWidth
            assert uvMap.shape[1] == gridHeight
            assert uvMap.shape[2] == 4
            assert uvMap.shape[3] == 2

        elif vertices.ndim == 3:
            assert vertices.shape[0] == self.__faceCount
            assert vertices.shape[1] == 4
            assert vertices.shape[2] == 3

            assert faces.ndim == 2
            assert faces.shape[0] == self.__faceCount
            assert faces.shape[1] == 4

            assert uvMap.ndim == 3
            assert uvMap.shape[0] == self.__faceCount
            assert uvMap.shape[1] == 4
            assert uvMap.shape[2] == 2

        else:
            raise ValueError("Unsupported input shape")

        # We will be working on this VAO
        glBindVertexArray(self.__vaoId)

        # Commodity variables to memorize the id of each VBO
        vertexVboId = self.__vboIds[0]
        uvVboId = self.__vboIds[1]
        indexVboId = self.__vboIds[2]

        # Copy vertices data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, vertexVboId)
        vertices = np.ascontiguousarray(vertices.flatten())
        glBufferSubData(GL_ARRAY_BUFFER, 0, 4 * len(vertices), vertices)

        # Copy UV coords data to GPU
        glBindBuffer(GL_ARRAY_BUFFER, uvVboId)
        uvMap = np.ascontiguousarray(uvMap.flatten())
        glBufferSubData(GL_ARRAY_BUFFER, 0, 4 * len(uvMap), uvMap)

        # Copy index data to GPU
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, indexVboId)
        faces = np.ascontiguousarray(faces.flatten())
        glBufferSubData(GL_ELEMENT_ARRAY_BUFFER, 0, 4 * len(faces), faces)

    def draw(self):
        glBindVertexArray(self.__vaoId)
        glBindTexture(GL_TEXTURE_2D, self.__textureId)
        glEnableVertexAttribArray(0)
        glEnableVertexAttribArray(1)
        glDrawElements(GL_QUADS, 4 * self.__renderedFaceCount, GL_UNSIGNED_INT, None)
        glDisableVertexAttribArray(0)
        glDisableVertexAttribArray(1)
        glBindVertexArray(0)

    def deleteObjects(self):
        glDeleteBuffers(3, self.__vboIds)
        glDeleteTextures(1, [self.__textureId])
        glDeleteVertexArrays(1, [self.__vaoId])
