import pygame
import numpy as np

from . import CharacterRenderer
from .. import surface2numpy, numpy2surface
from gui.text import TextStyle


class OutlineCharacterRenderer(CharacterRenderer):

    def __init__(self, style: TextStyle, characterRenderer: CharacterRenderer):
        self.__color = style.outlineColor
        self.__thickness = style.outlineThickness
        self.__characterRenderer = characterRenderer

    @property
    def tileDefaultWidth(self) -> int:
        return self.__characterRenderer.tileDefaultWidth + 2 * self.__thickness

    @property
    def tileDefaultHeight(self) -> int:
        return self.__characterRenderer.tileDefaultHeight + 2 * self.__thickness

    def getCharacterAdvance(self, character: str) -> int:
        return self.__characterRenderer.getCharacterAdvance(character) + 2 * self.__thickness

    def render(self, text: str) -> pygame.Surface:
        surface = self.__characterRenderer.render(text)

        # Create a larger surface
        size = (surface.get_width() + 2 * self.__thickness, surface.get_height() + 2 * self.__thickness)
        outlineSurface = pygame.Surface(size, flags=pygame.SRCALPHA)

        # Blit the character
        outlineSurface.blit(surface, (self.__thickness, self.__thickness))

        # Draw the outline
        outlineColor = self.__color + (255,)
        if True:  # Method with numpy (runs 50x faster on my computer)
            array = surface2numpy(outlineSurface)
            for step in range(self.__thickness):
                array = outlineStep(array, outlineColor)
            outlineSurface = numpy2surface(array)
        else:  # Method with pygame surfaces
            for step in range(self.__thickness):
                outlineSurface = outlineStepSurface(outlineSurface, outlineColor)

        return outlineSurface


def shiftXY(array: np.ndarray, directionX: int, directionY: int) -> np.ndarray:
    result = np.empty_like(array)
    height = result.shape[0]
    width = result.shape[1]

    if directionX > 0:
        inputRangeX = slice(directionX, width)
        outputRangeX = slice(0, width - directionX)
        fillRangeX = slice(width - directionX, width)
    elif directionX < 0:
        directionX = -directionX
        inputRangeX = slice(0, width - directionX)
        outputRangeX = slice(directionX, width)
        fillRangeX = slice(0, directionX)
    else:
        inputRangeX = slice(None)
        outputRangeX = slice(None)
        fillRangeX = None

    if directionY > 0:
        inputRangeY = slice(directionY, height)
        outputRangeY = slice(0, height - directionY)
        fillRangeY = slice(height - directionY, height)
    elif directionY < 0:
        directionY = -directionY
        inputRangeY = slice(0, height - directionY)
        outputRangeY = slice(directionY, height)
        fillRangeY = slice(0, directionY)
    else:
        inputRangeY = slice(None)
        outputRangeY = slice(None)
        fillRangeY = None

    result[outputRangeY, outputRangeX] = array[inputRangeY, inputRangeX]
    if fillRangeY is not None:
        result[fillRangeY, :] = 0
    if fillRangeX is not None:
        result[:, fillRangeX] = 0
    return result


def outlineStep(array: np.ndarray, outlineColor: (int, int, int, int)) -> np.ndarray:
    array = array.copy()
    alpha = array[..., 3]
    mask = shiftXY(alpha, 1, 0) | shiftXY(alpha, -1, 0)
    mask |= shiftXY(alpha, 0, 1) | shiftXY(alpha, 0, -1)
    mask |= shiftXY(alpha, -1, -1) | shiftXY(alpha, 1, 1)
    mask |= shiftXY(alpha, -1, 1) | shiftXY(alpha, 1, -1)
    mask &= ~alpha
    array[mask != 0] = outlineColor
    return array


def outlineStepSurface(surface: pygame.Surface, outlineColor: (int, int, int, int)) -> pygame.Surface:
    result = surface.copy()
    w = surface.get_width()
    h = surface.get_height()
    for y in range(h):
        for x in range(w):
            # Copy opaque pixels
            color = surface.get_at((x, y))
            if color[3] != 0:
                result.set_at((x, y), color)
                continue
            # If any pixel around the current one is opaque, then light it
            if (x > 0 and surface.get_at((x - 1, y))[3] != 0) \
                    or (x < w - 1 and surface.get_at((x + 1, y))[3] != 0) \
                    or (y > 0 and surface.get_at((x, y - 1))[3] != 0) \
                    or (y < h - 1 and surface.get_at((x, y + 1))[3] != 0) \
                    or (x > 0 and y > 0 and surface.get_at((x - 1, y - 1))[3] != 0) \
                    or (x > 0 and y < h - 1 and surface.get_at((x - 1, y + 1))[3] != 0) \
                    or (x < w - 1 and y > 0 and surface.get_at((x + 1, y - 1))[3] != 0) \
                    or (x < w - 1 and y < h - 1 and surface.get_at((x + 1, y + 1))[3] != 0):
                result.set_at((x, y), outlineColor)
    return result
