import random

from gui import GUIFacade
from gui.input import Keyboard, KeyEvent
from gui.text import TextStyle
from logic.command import CreateRegionCommand, SetGroundValueCommand, FillGroundValueCommand, SetWaterValueCommand, FillWaterValueCommand, \
    SetTreeValueCommand, FillTreeValueCommand, LoadStateCommand, SaveStateCommand, CharacterLocationCommand, CharacterCreateCommand
from state.Character import CHARACTER_CATEGORY_HERO
from state.Region import GROUND_GRASS, GROUND_SWAMP, GROUND_DIRT, GROUND_SAND, GROUND_DEAD_LAND, WATER_DARK, WATER_RIVER, WATER_SWAMP, WATER_CAVE, \
    WATER_SEA, TREE_LIGHT_GREEN, TREE_GREEN, TREE_ORANGE, TREE_DEAD, TREE_STUMP_LARGE, TREE_HERB_1, TREE_FLOWER_1, \
    TREE_MUSHROOM_1, TREE_NENUPHAR_1
from .DefaultGameMode import DefaultGameMode

BRUSH_GROUND = 0
BRUSH_WATER = 1
BRUSH_TREE = 2
BRUSH_CHARACTER = 3
BRUSH_MAX = 3

BRUSHES = [
    ("Ground", {
        "values": [
            ("Grass", GROUND_GRASS),
            ("Swamp", GROUND_SWAMP),
            ("Dirt", GROUND_DIRT),
            ("Sand", GROUND_SAND),
            ("Dead Land", GROUND_DEAD_LAND)
        ],
        "fill": FillGroundValueCommand,
        "set": SetGroundValueCommand
    }),
    ("Water", {
        "values": [
            ("Dark water", WATER_DARK),
            ("River", WATER_RIVER),
            ("Swamp", WATER_SWAMP),
            ("Sea", WATER_SEA),
            ("Cave", WATER_CAVE)
        ],
        "fill": FillWaterValueCommand,
        "set": SetWaterValueCommand
    }),
    ("Tree", {
        "values": [
            ("Tree light green", TREE_LIGHT_GREEN),
            ("Tree green", TREE_GREEN),
            ("Tree orange", TREE_ORANGE),
            ("Tree dead", TREE_DEAD),
            ("Tree stump", TREE_STUMP_LARGE),
            ("Herb", TREE_HERB_1),
            ("Flower", TREE_FLOWER_1),
            ("Mushroom", TREE_MUSHROOM_1),
            ("Nenuphar", TREE_NENUPHAR_1)
        ],
        "fill": FillTreeValueCommand,
        "set": SetTreeValueCommand
    }),
    ("Character", {
        "values": [
            ("Hero", CHARACTER_CATEGORY_HERO)
        ],
        "fill": None,
        "set": None
    })
]


class EditGameMode(DefaultGameMode):

    def __init__(self, gui: GUIFacade):
        super(EditGameMode, self).__init__(gui)

        self.__currentBrushMode = BRUSH_GROUND
        self.__currentBrush = GROUND_SWAMP

        self.__frontLayerGroup = gui.createLayerGroup()
        self.__framesLayer = self.__frontLayerGroup.createUILayer()
        self.__framesLayer.setTileset("assets/ui/ui.png")
        self.__framesLayer.setTileSize(self._tileSize, self._tileSize)
        self.__framesLayer.createFrameBox(
            self._tileSize,
            gui.screenHeight - self._tileSize * 8,
            gui.screenWidth - self._tileSize * 11,
            gui.screenHeight - self._tileSize,
        )
        style = TextStyle("assets/font/terminal-grotesque/terminal-grotesque.ttf", self._tileSize, (255, 255, 255))
        self.__textLayer = self.__frontLayerGroup.createTextLayer()
        self.__textLayer.setStyle(style)
        self.__textLayer.setMaxCharacterCount(500)
        self.__textLayer.setText(
            self._tileSize * 1.5,
            gui.screenHeight - self._tileSize * 7.5,
            "<u><b>Commands</b></u>\n"
            "<b>Left click:</b> Paint\n"
            "<b>Middle click:</b> Fill\n"
            "<b>Right click:</b> Remove\n"
            "<b>F1/F2:</b> Load/Save\n"
            "<b>F3:</b> Change brush\n"
        )

        self.__brushTextLayer = self.__frontLayerGroup.createTextLayer()
        self.__brushTextLayer.setStyle(style)
        self.__brushTextLayer.setMaxCharacterCount(500)
        self.__brushTextLayer2 = self.__frontLayerGroup.createTextLayer()
        self.__brushTextLayer2.setStyle(style)
        self.__brushTextLayer2.setMaxCharacterCount(500)
        self.updateBrushText()

        self._gui.layersChanged()

    def updateBrushText(self):

        brush = BRUSHES[self.__currentBrushMode]
        text1 = "Current brush: {}\n".format(brush[0])
        text2 = "\n"
        for i, pair in enumerate(brush[1]["values"]):
            name = pair[0]
            text = "<b>{}</b>: {}\n".format(i + 1, name)
            if i < 5:
                text1 += text
            else:
                text2 += text

        self.__brushTextLayer.setText(
            self._tileSize * 10.5,
            self._gui.screenHeight - self._tileSize * 7.5,
            text1
        )
        self.__brushTextLayer2.setText(
            self._tileSize * 20.5,
            self._gui.screenHeight - self._tileSize * 7.5,
            text2
        )

    #############################################################
    # Mode
    #############################################################

    def init(self):
        self._logic.addCommand(CreateRegionCommand(self._state, 'test', 64, 48))

    def handleInputs(self) -> bool:
        # Mouse
        mouse = self._gui.mouse
        mouseX = (mouse.x + self._viewX) // self._tileSize
        mouseY = (mouse.y + self._viewY) // self._tileSize

        if mouse.button1:
            if self.__currentBrushMode == BRUSH_CHARACTER:
                characterIndex = self.__currentBrush - 1
                if self._currentRegion.findCharacter(characterIndex) is None:
                    self._logic.addCommand(CharacterCreateCommand(self._state, self._currentRegionName,
                                           characterIndex, characterIndex + 1, mouseX, mouseY))
                else:
                    self._logic.addCommand(CharacterLocationCommand(self._state, self._currentRegionName,
                                           characterIndex, mouseX, mouseY))
            else:
                brush = self.__currentBrush
                if self.__currentBrushMode == BRUSH_TREE:
                    if brush == TREE_FLOWER_1 or brush == TREE_HERB_1:
                        brush += random.randint(0, 3)
                    elif brush == TREE_MUSHROOM_1 or brush == TREE_NENUPHAR_1:
                        brush += random.randint(0, 1)
                command = BRUSHES[self.__currentBrushMode][1]["set"]
                self._logic.addCommand(command(self._state, self._currentRegionName, mouseX, mouseY, brush))
                return True
        if mouse.button3:
            if self.__currentBrushMode == BRUSH_GROUND:
                brush = GROUND_GRASS
            else:
                brush = 0
            command = BRUSHES[self.__currentBrushMode][1]["set"]
            if command is not None:
                self._logic.addCommand(command(self._state, self._currentRegionName, mouseX, mouseY, brush))
                return True
        if mouse.button2:
            if self.__currentBrushMode == BRUSH_GROUND:
                value = self._currentRegion.getGroundValue(mouseX, mouseY)
            elif self.__currentBrushMode == BRUSH_WATER:
                value = self._currentRegion.getWaterValue(mouseX, mouseY)
            elif self.__currentBrushMode == BRUSH_TREE:
                value = self._currentRegion.getTreeValue(mouseX, mouseY)
            else:
                value = -1
            if value >= 0:
                command = BRUSHES[self.__currentBrushMode][1]["fill"]
                if command is not None:
                    self._logic.addCommand(command(self._state, self._currentRegionName, mouseX, mouseY, value, self.__currentBrush))
                    return True

        # Keyboard
        keyboard = self._gui.keyboard
        shiftX = 0
        shiftY = 0
        if keyboard.isKeyPressed(Keyboard.K_LEFT):
            shiftX -= self._tileSize
        elif keyboard.isKeyPressed(Keyboard.K_RIGHT):
            shiftX += self._tileSize
        elif keyboard.isKeyPressed(Keyboard.K_UP):
            shiftY -= self._tileSize
        elif keyboard.isKeyPressed(Keyboard.K_DOWN):
            shiftY += self._tileSize
        if shiftX != 0 or shiftY != 0:
            self._viewX += shiftX
            self._viewY += shiftY
            self._worldLayerGroup.setTranslation(self._viewX, self._viewY)
            return True

        for event in keyboard.keyEvents:
            if event.type != KeyEvent.KEYDOWN:
                continue
            if event.key == Keyboard.K_F1:
                self._logic.addCommand(LoadStateCommand(self, "level.bin"))
                keyboard.clearKeyEvents()
                return True
            elif event.key == Keyboard.K_F2:
                self._logic.addCommand(SaveStateCommand(self._state, "level.bin"))
                keyboard.clearKeyEvents()
                return True
            elif event.key == Keyboard.K_F3:
                self.__currentBrushMode += 1
                if self.__currentBrushMode > BRUSH_MAX:
                    self.__currentBrushMode = 0
                self.__currentBrush = 1
                self.updateBrushText()
                keyboard.clearKeyEvents()
                return True
            elif Keyboard.K_0 <= event.key <= Keyboard.K_9:
                key2index = [10, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
                index = key2index[event.key - Keyboard.K_0]
                values = BRUSHES[self.__currentBrushMode][1]["values"]
                if index < len(values):
                    self.__currentBrush = values[index][1]
                    keyboard.clearKeyEvents()
                    return True

        return False










