from typing import Union, List, Dict

from gui import GUIFacade
from gui.layer import CharactersLayer, LayerGroup
from state import Region, Character, State, Direction
from .Renderer import Renderer
from .animation import CharacterAnimation


class CharactersRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, layerIndex: int, tileSize: int):
        super(CharactersRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__layerIndex = layerIndex
        self.__tileSize = tileSize  # type: int
        self.__layer = None  # type: Union[None, CharactersLayer]
        self.__animations = {}  # type: Dict[int, CharacterAnimation]

    def updateAll(self, region: Union[None, Region]):
        # Case where we remove the region
        if region is None:
            if self.__layer is not None:
                self.__layerGroup.removeLayer(self.__layer)
                self.__layer = None
            return

        characters = region.characters

        # Create the layer
        self.__layer = self.__layerGroup.createCharactersLayer(self.__layerIndex)
        self.__layer.setTileset("assets/tile/characters.png")
        self.__layer.setTileSize(self.__tileSize, self.__tileSize)
        self.__layer.setGridSize(region.width, region.height)
        self.__layer.setCharacterCount(len(characters))
        self.__animations = {}

        for index in range(len(characters)):
            self.updateCharacter(index, characters[index])

    def updateCharacter(self, index: int, character: Character):
        if index >= self.__layer.getCharacterCount():
            self.__layer.setCharacterCount(index + 1)

        if character is not None:
            self.__animations[index] = CharacterAnimation(self.__layer, index, character)
        else:
            del self.__animations[index]

    def updateAnimations(self, time: float, epochFrame: float):
        if self.__layer is None:
            return
        for animation in self.__animations.values():
            animation.update(time, epochFrame)

    def getCharacterLocation(self, characterIndex: int) -> (float, float):
        if characterIndex in self.__animations:
            return self.__animations[characterIndex].getLocation()
        return -1, -1

