from typing import Union

from gui import GUIFacade
from gui.layer import LayerGroup
from state import Region, State, Character
from .CharactersRenderer import CharactersRenderer
from .GroundRenderer import GroundRenderer
from .Renderer import Renderer
from .TreeRenderer import TreeRenderer
from .WaterRenderer import WaterRenderer


class RegionRenderer(Renderer):

    def __init__(self, gui: GUIFacade, layerGroup: LayerGroup, tileSize: int):
        super(RegionRenderer, self).__init__(gui)
        self.__layerGroup = layerGroup
        self.__tileSize = tileSize
        self.__regionWidth = 0
        self.__regionHeight = 0
        self.__charactersRenderer = CharactersRenderer(gui, layerGroup, 3, tileSize)
        self.__renderers = [
            GroundRenderer(gui, layerGroup, 0, tileSize),
            WaterRenderer(gui, layerGroup, 1, tileSize),
            TreeRenderer(gui, layerGroup, 2, tileSize),
            self.__charactersRenderer
        ]

    def updateViewOnCharacter(self, characterIndex: int):
        worldX, worldY = self.__charactersRenderer.getCharacterLocation(characterIndex)
        if worldX < 0 or worldY < 0:
            return
        worldX *= self.__tileSize
        worldY *= self.__tileSize
        viewX, viewY = self.__layerGroup.getTranslation()
        screenX = worldX - viewX
        screenY = worldY - viewY

        # Move view if character is too close to the border of the screen
        margin = self._gui.screenWidth * 0.3
        if screenX < margin:
            viewX = worldX - margin
        elif screenX > self._gui.screenWidth - margin:
            viewX = worldX - self._gui.screenWidth + margin
        margin = self._gui.screenHeight * 0.3
        if screenY < margin:
            viewY = worldY - margin
        elif screenY > self._gui.screenHeight - margin:
            viewY = worldY - self._gui.screenHeight + margin

        # Only show world in the window
        if viewX < 0:
            viewX = 0
        if viewY < 0:
            viewY = 0
        maxX = self.__regionWidth * self.__tileSize - self._gui.screenWidth
        if viewX >= maxX:
            viewX = maxX
        maxY = self.__regionHeight * self.__tileSize - self._gui.screenHeight
        if viewY >= maxY:
            viewY = maxY

        self.__layerGroup.setTranslation(viewX, viewY)

    def updateAll(self, region: Union[None, Region]):
        if region is not None:
            self.__regionWidth = region.width
            self.__regionHeight = region.height
        for renderer in self.__renderers:
            renderer.updateAll(region)

    def updateTile(self, region: Region, x: int, y: int):
        for renderer in self.__renderers:
            renderer.updateTile(region, x, y)

    def updateCharacter(self, index: int, character: Character):
        for renderer in self.__renderers:
            renderer.updateCharacter(index, character)

    def updateAnimations(self, time: float, epochFrame: float):
        for renderer in self.__renderers:
            renderer.updateAnimations(time, epochFrame)
