from __future__ import annotations
from typing import Dict, List, Union

from .StateListener import StateListener
from .Region import Region

class State:

    def __init__(self):

        self.__listeners = []  # type: List[StateListener]
        self.__regions = {}  # type: Dict[str, Region]

    # Listeners

    def addListener(self, listener: StateListener):
        if listener not in self.__listeners:
            self.__listeners.append(listener)

    def removeListener(self, listener: StateListener):
        if listener in self.__listeners:
            self.__listeners.remove(listener)

    def notifyCurrentRegionChanged(self, regionName: str):
        for listener in self.__listeners:
            listener.currentRegionChanged(self, regionName)

    def notifyRegionCellChanged(self, regionName: str, x: int, y: int):
        for listener in self.__listeners:
            listener.regionCellChanged(self, regionName, x, y)

    def notifyCharacterChanged(self, state: State, regionName: str, characterIndex: int):
        for listener in self.__listeners:
            listener.characterChanged(state, regionName, characterIndex)

    # Data

    @property
    def regions(self) -> Dict[str, Region]:
        return self.__regions

    @regions.setter
    def regions(self, regions: Dict[str, Region]):
        self.__regions = regions

    def createRegion(self, name: str, width: int, height: int) -> Region:
        if name in self.__regions:
            raise ValueError("Region {} already exists".format(name))
        region = Region(width, height)
        self.__regions[name] = region
        return region

    def findRegion(self, name: str) -> Union[None, Region]:
        if name not in self.__regions:
            return None
        return self.__regions[name]

    def getRegionNames(self) -> List[str]:
        return list(self.__regions.keys())

    def getRegion(self, name: str) -> Region:
        if name not in self.__regions:
            raise ValueError("There is no region {}".format(name))
        return self.__regions[name]

