from __future__ import annotations

import io
import struct
import numpy as np
from typing import Union, Any, Dict, List

from .ObjectSerializer import ObjectSerializer


class IO:

    def __init__(self, stream: Union[None, io.BytesIO] = None):
        self.__stream = stream
        self.__types = {}  # type: Dict[str, ObjectSerializer]

    @property
    def stream(self):
        return self.__stream

    def createStream(self):
        self.__stream = io.BytesIO()

    def loadFile(self, fileName: str):
        with open(fileName, "rb") as f:
            self.__stream = io.BytesIO(f.read())

    def saveFile(self, fileName: str):
        self.__stream.seek(0)
        with open(fileName, "wb") as f:
            f.write(self.__stream.read())

    #######################################################################
    # Headers

    def writeFileHeader(self, sign: str, version: int):
        self.writeString(sign)
        self.writeUInt64(version)

    def readFileHeader(self, sign: str) -> int:
        assert self.readString() == sign
        return self.readUInt64()

    def writeBlockHeader(self, code: int):
        assert 0 <= code <= 255
        data = code.to_bytes(1, byteorder='little', signed=False)
        self.__stream.write(data)

    def readBlockHeader(self) -> int:
        data = self.__stream.read(1)
        return int.from_bytes(data, byteorder='little', signed=False)

    #######################################################################
    # String

    def writeString(self, s: str):
        self.writeBlockHeader(self.BLOCK_STRING)
        data = s.encode('utf-8')
        length = len(data)
        assert 0 <= length <= 65535
        self.__stream.write(length.to_bytes(2, byteorder='little', signed=False))
        self.__stream.write(data)

    def readString(self) -> str:
        assert self.readBlockHeader() == self.BLOCK_STRING
        data = self.__stream.read(2)
        length = int.from_bytes(data, byteorder='little', signed=False)
        data = self.__stream.read(length)
        return data.decode('utf-8')

    #######################################################################
    # Int8

    def writeUInt8(self, i: int):
        assert 0 <= i <= 255
        self.writeBlockHeader(self.BLOCK_UINT8)
        data = i.to_bytes(1, byteorder='little', signed=False)
        self.__stream.write(data)

    def readUInt8(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_UINT8
        data = self.__stream.read(1)
        return int.from_bytes(data, byteorder='little', signed=False)

    def writeInt8(self, i: int):
        assert -128 <= i <= 127
        self.writeBlockHeader(self.BLOCK_INT8)
        data = i.to_bytes(1, byteorder='little', signed=True)
        self.__stream.write(data)

    def readInt8(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_INT8
        data = self.__stream.read(1)
        return int.from_bytes(data, byteorder='little', signed=True)

    #######################################################################
    # Int16

    def writeUInt16(self, i: int):
        assert 0 <= i <= 2 ** 16 - 1
        self.writeBlockHeader(self.BLOCK_UINT16)
        data = i.to_bytes(2, byteorder='little', signed=False)
        self.__stream.write(data)

    def readUInt16(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_UINT16
        data = self.__stream.read(2)
        return int.from_bytes(data, byteorder='little', signed=False)

    def writeInt16(self, i: int):
        assert -2 ** 15 <= i <= 2 ** 15 - 1
        self.writeBlockHeader(self.BLOCK_INT16)
        data = i.to_bytes(2, byteorder='little', signed=True)
        self.__stream.write(data)

    def readInt16(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_INT16
        data = self.__stream.read(2)
        return int.from_bytes(data, byteorder='little', signed=True)

    #######################################################################
    # Int32

    def writeUInt32(self, i: int):
        assert 0 <= i <= 2 ** 32 - 1
        self.writeBlockHeader(self.BLOCK_UINT32)
        data = i.to_bytes(4, byteorder='little', signed=False)
        self.__stream.write(data)

    def readUInt32(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_UINT32
        data = self.__stream.read(4)
        return int.from_bytes(data, byteorder='little', signed=False)

    def writeInt32(self, i: int):
        assert -2 ** 31 <= i <= 2 ** 31 - 1
        self.writeBlockHeader(self.BLOCK_INT32)
        data = i.to_bytes(4, byteorder='little', signed=True)
        self.__stream.write(data)

    def readInt32(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_INT32
        data = self.__stream.read(4)
        return int.from_bytes(data, byteorder='little', signed=True)

    #######################################################################
    # Int64

    def writeUInt64(self, i: int):
        assert 0 <= i <= 2 ** 64 - 1
        self.writeBlockHeader(self.BLOCK_UINT64)
        data = i.to_bytes(8, byteorder='little', signed=False)
        self.__stream.write(data)

    def readUInt64(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_UINT64
        data = self.__stream.read(8)
        return int.from_bytes(data, byteorder='little', signed=False)

    def writeInt64(self, i: int):
        assert -2 ** 63 <= i <= 2 ** 63 - 1
        self.writeBlockHeader(self.BLOCK_INT64)
        data = i.to_bytes(8, byteorder='little', signed=True)
        self.__stream.write(data)

    def readInt64(self) -> int:
        assert self.readBlockHeader() == self.BLOCK_INT64
        data = self.__stream.read(8)
        return int.from_bytes(data, byteorder='little', signed=True)

    #######################################################################
    # Float

    def writeFloat64(self, f: float):
        self.writeBlockHeader(self.BLOCK_FLOAT64)
        data = struct.pack("<d", f)
        self.__stream.write(data)

    def readFloat64(self) -> float:
        assert self.readBlockHeader() == self.BLOCK_FLOAT64
        data = self.__stream.read(8)
        return struct.unpack("<d", data)[0]

    #######################################################################
    # Numpy arrays

    def writeNumpyArray(self, a: np.ndarray):
        self.writeBlockHeader(self.BLOCK_NUMPY_ARRAY)
        # Shape
        ndim = a.ndim
        data = ndim.to_bytes(8, byteorder='little', signed=False)
        self.__stream.write(data)
        for dim in range(ndim):
            data = a.shape[dim].to_bytes(8, byteorder='little', signed=False)
            self.__stream.write(data)
        # Data type
        typeCode = 0  # type: int
        if a.dtype == np.uint8:
            typeCode = self.BLOCK_UINT8
        elif a.dtype == np.int8:
            typeCode = self.BLOCK_INT8
        elif a.dtype == np.uint16:
            typeCode = self.BLOCK_UINT16
        elif a.dtype == np.int16:
            typeCode = self.BLOCK_INT16
        elif a.dtype == np.uint32:
            typeCode = self.BLOCK_UINT32
        elif a.dtype == np.int32:
            typeCode = self.BLOCK_INT32
        elif a.dtype == np.uint64:
            typeCode = self.BLOCK_UINT64
        elif a.dtype == np.int64:
            typeCode = self.BLOCK_INT64
        elif a.dtype == np.float32:
            typeCode = self.BLOCK_FLOAT32
        elif a.dtype == np.float64:
            typeCode = self.BLOCK_FLOAT64
        else:
            raise ValueError("Unsupported numpy array data type {}".format(a.dtype))
        data = typeCode.to_bytes(8, byteorder='little', signed=False)
        self.__stream.write(data)

        # Content
        data = a.tobytes('C')
        dataSize = len(data).to_bytes(8, byteorder='little', signed=False)
        self.__stream.write(dataSize)
        self.__stream.write(data)

    def readNumpyArray(self) -> np.ndarray:
        assert self.readBlockHeader() == self.BLOCK_NUMPY_ARRAY
        # Shape
        data = self.__stream.read(8)
        ndim = int.from_bytes(data, byteorder='little', signed=False)
        shape = []
        for dim in range(ndim):
            data = self.__stream.read(8)
            shape.append(int.from_bytes(data, byteorder='little', signed=False))
        # Data type
        data = self.__stream.read(8)
        typeCode = int.from_bytes(data, byteorder='little', signed=False)
        if typeCode == self.BLOCK_UINT8:
            dtype = np.uint8
        elif typeCode == self.BLOCK_INT8:
            dtype = np.int8
        elif typeCode == self.BLOCK_UINT16:
            dtype = np.uint16
        elif typeCode == self.BLOCK_INT16:
            dtype = np.int16
        elif typeCode == self.BLOCK_UINT32:
            dtype = np.uint32
        elif typeCode == self.BLOCK_INT32:
            dtype = np.int32
        elif typeCode == self.BLOCK_UINT64:
            dtype = np.uint64
        elif typeCode == self.BLOCK_INT64:
            dtype = np.int64
        elif typeCode == self.BLOCK_FLOAT32:
            dtype = np.float32
        elif typeCode == self.BLOCK_FLOAT64:
            dtype = np.float64
        else:
            raise RuntimeError("Unsupported numpy array data type {}".format(typeCode))
        # Create array
        data = self.__stream.read(8)
        dataSize = int.from_bytes(data, byteorder='little', signed=False)
        data = self.__stream.read(dataSize)
        a = np.frombuffer(data, dtype=dtype).reshape(shape)
        return a.copy()

    #######################################################################
    # Object

    def registerObjectSerializer(self, t: str, serializer: ObjectSerializer):
        self.__types[t] = serializer

    def writeObject(self, o: Any):
        t = type(o).__name__
        if t not in self.__types:
            raise ValueError("Type {} is not registered".format(t))
        self.writeBlockHeader(self.BLOCK_OBJECT)
        self.writeString(t)
        serializer = self.__types[t]
        serializer.write(self, o)

    def readObject(self, className: Union[None, str] = None) -> Any:
        assert self.readBlockHeader() == self.BLOCK_OBJECT
        t = self.readString()
        if t not in self.__types:
            raise ValueError("Type {} is not registered".format(t))
        serializer = self.__types[t]
        o = serializer.read(self)
        if className is not None:
            if o.__class__.__name__ != className:
                raise ValueError("Invalid object type: expect '{}' but got '{}'".format(className, object.__class__.__name__))
        return o

    def writeObjectList(self, array: List):
        self.writeBlockHeader(self.BLOCK_OBJECT_LIST)
        self.writeInt64(len(array))
        for value in array:
            self.writeObject(value)

    def readObjectList(self, className: Union[None, str] = None) -> List:
        assert self.readBlockHeader() == self.BLOCK_OBJECT_LIST
        array = []
        valueCount = self.readInt64()
        for valueIndex in range(valueCount):
            array.append(self.readObject(className))
        return array

    def writeStringObjectDict(self, dict: Dict[str]):
        self.writeBlockHeader(self.BLOCK_OBJECT_DICT)
        self.writeInt64(len(dict))
        for name, value in dict.items():
            self.writeString(name)
            self.writeObject(value)

    def readStringObjectDict(self, className: Union[None, str] = None) -> Dict[str]:
        assert self.readBlockHeader() == self.BLOCK_OBJECT_DICT
        d = {}
        valueCount = self.readInt64()
        for valueIndex in range(valueCount):
            name = self.readString()
            value = self.readObject(className)
            d[name] = value
        return d

    #######################################################################
    # Block codes

    BLOCK_STRING = 0x01
    BLOCK_OBJECT = 0x02
    BLOCK_OBJECT_DICT = 0x03
    BLOCK_OBJECT_LIST = 0x04

    BLOCK_UINT8 = 0x10
    BLOCK_INT8 = 0x11
    BLOCK_UINT16 = 0x12
    BLOCK_INT16 = 0x13
    BLOCK_UINT32 = 0x14
    BLOCK_INT32 = 0x15
    BLOCK_UINT64 = 0x16
    BLOCK_INT64 = 0x17

    BLOCK_FLOAT32 = 0x20
    BLOCK_FLOAT64 = 0x21
    
    BLOCK_NUMPY_ARRAY = 0x30

    #######################################################################
    # Tests

    def tests(self):
        self.createStream()
        self.writeString("Méssàge wïth spéciàl çharacters")
        self.writeInt8(-12)
        self.writeUInt8(23)
        self.writeInt16(-0x1234)
        self.writeUInt16(0x1234)
        self.writeInt32(-0x12345678)
        self.writeUInt32(0x12345678)
        self.writeInt64(-0x1234567812345678)
        self.writeUInt64(0x1234567812345678)
        self.writeFloat64(123.45678)

        uint8np = np.random.randint(0, 255, [3, 4], dtype=np.uint8)
        self.writeNumpyArray(uint8np)
        int8np = np.random.randint(-128, 127, [4, 12], dtype=np.int8)
        self.writeNumpyArray(int8np)
        
        uint16np = np.random.randint(0, 2**16 - 1, [3, 4], dtype=np.uint16)
        self.writeNumpyArray(uint16np)
        int16np = np.random.randint(-2**15, 2**15 - 1, [4, 12], dtype=np.int16)
        self.writeNumpyArray(int16np)
        
        uint32np = np.random.randint(0, 2**32 - 1, [3, 4], dtype=np.uint32)
        self.writeNumpyArray(uint32np)
        int32np = np.random.randint(-2**31, 2**31 - 1, [4, 12], dtype=np.int32)
        self.writeNumpyArray(int32np)
        
        uint64np = np.random.randint(0, 2**64 - 1, [3, 4], dtype=np.uint64)
        self.writeNumpyArray(uint64np)
        int64np = np.random.randint(-2**63, 2**63 - 1, [4, 12], dtype=np.int64)
        self.writeNumpyArray(int64np)

        float32np = np.random.rand(3, 4).astype(np.float32) * 123
        self.writeNumpyArray(float32np)
        float64np = np.random.rand(5, 12).astype(np.float64) * 123
        self.writeNumpyArray(float64np)

        class User:
            def __init__(self, name: str, age: int):
                self.name = name
                self.age = age

            def __eq__(self, other: User):
                return self.name == other.name and self.age == other.age

            def __repr__(self):
                return "{}:{}".format(self.name, self.age)

        class UserSerializer(ObjectSerializer):
            def write(self, io: IO, user: User):
                io.writeString(user.name)
                io.writeUInt8(user.age)

            def read(self, io: IO) -> User:
                name = io.readString()
                age = io.readUInt8()
                return User(name, age)

        self.registerObjectSerializer("User", UserSerializer())
        self.writeObject(User("Peter", 23))

        users = {
            "Story": User("Peter", 23),
            "Graphics": User("Sarah", 21),
            "Code": User("Paul", 37)
        }
        self.writeStringObjectDict(users)

        array = [
            User("Peter", 23),
            User("Sarah", 21),
            User("Paul", 37)
        ]
        self.writeObjectList(array)

        ####################################################################
        # Read

        self.__stream.seek(0)
        assert self.readString() == "Méssàge wïth spéciàl çharacters"
        assert self.readInt8() == -12
        assert self.readUInt8() == 23
        assert self.readInt16() == -0x1234
        assert self.readUInt16() == 0x1234
        assert self.readInt32() == -0x12345678
        assert self.readUInt32() == 0x12345678
        assert self.readInt64() == -0x1234567812345678
        assert self.readUInt64() == 0x1234567812345678
        assert self.readFloat64() == 123.45678

        a = self.readNumpyArray()
        assert a.shape == uint8np.shape
        assert a.dtype == uint8np.dtype
        assert (a == uint8np).all()
        a = self.readNumpyArray()
        assert a.shape == int8np.shape
        assert a.dtype == int8np.dtype
        assert (a == int8np).all()
        
        a = self.readNumpyArray()
        assert a.shape == uint16np.shape
        assert a.dtype == uint16np.dtype
        assert (a == uint16np).all()
        a = self.readNumpyArray()
        assert a.shape == int16np.shape
        assert a.dtype == int16np.dtype
        assert (a == int16np).all()

        a = self.readNumpyArray()
        assert a.shape == uint32np.shape
        assert a.dtype == uint32np.dtype
        assert (a == uint32np).all()
        a = self.readNumpyArray()
        assert a.shape == int32np.shape
        assert a.dtype == int32np.dtype
        assert (a == int32np).all()

        a = self.readNumpyArray()
        assert a.shape == uint64np.shape
        assert a.dtype == uint64np.dtype
        assert (a == uint64np).all()
        a = self.readNumpyArray()
        assert a.shape == int64np.shape
        assert a.dtype == int64np.dtype
        assert (a == int64np).all()

        a = self.readNumpyArray()
        assert a.shape == float32np.shape
        assert a.dtype == float32np.dtype
        assert (a == float32np).all()
        a = self.readNumpyArray()
        assert a.shape == float64np.shape
        assert a.dtype == float64np.dtype
        assert (a == float64np).all()

        user = self.readObject("User")
        assert type(user).__name__ == "User"
        assert isinstance(user, User)
        assert user.name == "Peter"
        assert user.age == 23

        users2 = self.readStringObjectDict("User")
        assert len(users) == len(users2)
        for name, value in users.items():
            assert name in users2
            assert users2[name] == value
        for name, value in users2.items():
            assert name in users
            assert users[name] == value

        array2 = self.readObjectList("User")
        assert len(array) == len(array2)
        for i in range(len(array2)):
            assert array[i] == array2[i]
